;; Generated package description from mastodon.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "mastodon" "2.0.16.0.20260406.85635" "Client for fediverse services using the Mastodon API" '((emacs "29.1") (persist "0.8") (tp "0.8")) :commit "5bba23045efda9f63c36ac431bec8f318a55e76a" :authors '(("Johnson Denen" . "johnson.denen@gmail.com") ("Marty Hiatt" . "mousebot@disroot.org")) :maintainer '("Marty Hiatt" . "mousebot@disroot.org") :url "https://codeberg.org/martianh/mastodon.el")
